use alloc_counter::{count_alloc, AllocCounterSystem};
use anyhow::Result;

#[global_allocator]
static A: AllocCounterSystem = AllocCounterSystem;

static INPUT: &str = include_str!("../../input02");

fn main() -> Result<()> {
    println!("validating...");
    let (cnt, _) = count_alloc(|| {
        let cnt = INPUT.lines().filter_map(validate_1).count();
        println!("part1: {} results are valid", cnt);
        let cnt = INPUT.lines().filter_map(validate_2).count();
        println!("part2: {} results are valid", cnt);
    });
    println!("{:?} allocations", cnt);
    Ok(())
}

fn validate_1(input: &str) -> Option<()> {
    let mut input = Splitter::from(input);
    let min = input
        .get_before('-')?
        .parse::<usize>()
        .expect("invalid input");
    input.skip();
    let max = input
        .get_before(' ')?
        .parse::<usize>()
        .expect("invalid input");
    input.skip();
    let c = input.get_before(':')?.chars().next()?;
    input.skip();
    input.skip();
    let input = &*input;

    let count = input.chars().filter(|&iter_c| iter_c == c).count();
    if count >= min && count <= max {
        return Some(());
    }
    None
}

fn validate_2(input: &str) -> Option<()> {
    let mut splitter = Splitter::from(input);
    let correct = splitter
        .get_before('-')
        .expect("invalid input")
        .parse::<usize>()
        .expect("invalid input");
    splitter.skip();
    let incorrect = splitter
        .get_before(' ')
        .expect("invalid input")
        .parse::<usize>()
        .expect("invalid input");
    splitter.skip();
    let c = splitter
        .get_before(':')
        .expect("invalid input")
        .chars()
        .next()
        .expect("invalid input");
    splitter.skip();
    splitter.skip();
    let splitter = &*splitter;

    let a = splitter.chars().nth(correct - 1).expect("invalid input") == c;
    let b = splitter.chars().nth(incorrect - 1).expect("invalid input") == c;
    if a as usize + b as usize == 1 {
        return Some(());
    }
    None
}

struct Splitter<'a> {
    inner: Option<&'a str>,
}

impl<'a> From<&'a str> for Splitter<'a> {
    fn from(inner: &'a str) -> Self {
        Splitter { inner: Some(inner) }
    }
}

impl<'a> std::ops::Deref for Splitter<'a> {
    type Target = str;
    fn deref(&self) -> &Self::Target {
        match self.inner {
            Some(inner) => inner,
            None => "",
        }
    }
}

impl<'a> Splitter<'a> {
    fn get_before(&mut self, r: char) -> Option<&'a str> {
        let inner = self.inner?;
        if let Some(pos) = inner.find(r) {
            let (before, after) = inner.split_at(pos);
            self.inner = Some(after);
            Some(before)
        } else {
            self.inner = None;
            Some(inner)
        }
    }
    fn skip(&mut self) {
        self.inner = match self.inner {
            Some(inner) => Some(&inner[1..]),
            None => None,
        };
    }
}
